/* File: parselabel.l
 * 
 * This file is part of XSCHEM,
 * a schematic capture and Spice/Vhdl/Verilog netlisting tool for circuit 
 * simulation.
 * Copyright (C) 1998-2020 Stefan Frederik Schippers
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */


%option never-interactive
%option noyywrap
%option noinput
%option nounput
%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CAD_SPICE_NETLIST 1

extern int netlist_type;
int dbg_var = 0;
extern FILE *errfp;
extern void my_strdup(int id, char **dest, char *src);
extern void my_free(void *ptr);
extern char bus_replacement_char[];
typedef struct          /* used in expandlabel.y */
{
 char *str;             /* label name */
 int m;                 /* label multiplicity, number of wires */
} Stringptr;

/* commented 20170412, fixes problems with older bison/flex versions */
/* #define YYPARSE_PARAM */
#include "expandlabel.h" /* Bison header file */

Stringptr dest_string={NULL,0};  /*19102004; */

static int bracket=0;
extern int yyparse(void) ;


void strreplace(char s[], char chr, char repl_chr)
{
     int i=0;
     while(s[i]!='\0')
     {
           if(s[i]==chr)
           {
               s[i]=repl_chr;
           }  
           i++; 
     }
}

const char *expandlabel(const char *s, int *m)
{
 YY_BUFFER_STATE buf;

 my_free(&dest_string.str);     /* 30102004  delete 'memory' of previous execution */
 dest_string.str=NULL;          /* 30102004 */

 if(dbg_var >=3) fprintf(errfp, "expandlabel(): entering\n");
 buf=yy_scan_string(s);
 bracket=0;
 yyparse();  /* 20140108 */
 yy_delete_buffer(buf);
 if(dbg_var >=3) fprintf(errfp, "expandlabel(): returning %s  from %s mult=%d\n",dest_string.str, s, dest_string.m);
 if(dest_string.str)
    *m = dest_string.m;
 else
    *m=-1;
 if(dest_string.str) {
   if(netlist_type == CAD_SPICE_NETLIST && bus_replacement_char[0]) {
     strreplace(dest_string.str, '[', bus_replacement_char[0]);
     strreplace(dest_string.str, ']', bus_replacement_char[1]);
   }
   return dest_string.str;
 } else {
   return s;
 }
}


%}
/*
   Lexical analyzer
*/

%x label
%x next_to_alias
%x rest

%%
^(alias|ipin|iopin|opin)[+ \n]+[^+\n ]+/[\n +]+ { 
                          yylval.ptr.str=NULL; /*19102004 */
       			  /* these things are freed after use in expandlabel.y */
			  my_strdup(298, &yylval.ptr.str, yytext);
                          BEGIN(next_to_alias);
			  if(dbg_var >=3) fprintf(errfp, "parselabel: alias, yytext = %s\n", yytext);
                          return B_LINE;
                        }
<next_to_alias>[\n +]+  {       /* get all white space and return a single separator */
                          BEGIN(label);
			  if(dbg_var >=3) fprintf(errfp, "parselabel: next_to_alias, yytext = %s\n", yytext);
                          yylval.val=yytext[0];
                          return B_CAR;
                        }
^[^*]                   {
                          if(dbg_var>=3) fprintf(errfp, "yylex(): matched: ^[^*] |%s|, push back\n",yytext);
                          yyless(0);    /* push back to input */
                          BEGIN(label); /* we know that a label follows. */
                                       /* so go and process it. */
                        }
^\*.*                   {              /* a comment, return as LINE token */
                          yylval.ptr.str=NULL; /*19102004 */
                          if(dbg_var>=3) fprintf(errfp, "yylex(): B_LINE: %s\n",yytext);
                          my_strdup(299, &yylval.ptr.str, yytext); /* these things are freed after use in expandlabel.y */
                          return B_LINE;
                        }
<label>{

[ \n]+                  {       /* this ends the <label> start condition */
                          BEGIN(rest);
                          yylval.val=yytext[0];
                          if(dbg_var>=3) fprintf(errfp, "yylex(): B_CAR: %s\n",yytext);
                          return B_CAR;
                        }
\.\.                    {
                          if(dbg_var>=3) fprintf(errfp, "parselabel(): B_DOUBLEDOT: %s\n", yytext);
                          return B_DOUBLEDOT;
                        }
[0-9]+                  {
                          sscanf(yytext, "%d",&yylval.val);
                          if(bracket) {
                            if(dbg_var>=3) fprintf(errfp, "parselabel(): B_IDXNUM: %s\n", yytext);
                            return B_IDXNUM;
                          }
                          else {
                            if(dbg_var>=3) fprintf(errfp, "parselabel(): B_NUM: %s\n", yytext);
                            return B_NUM;
                          }
                        }
([-+=_a-zA-Z][-\\/:.=_+a-zA-Z0-9]*)|([-+.=_a-zA-Z][-\\/:=_+a-zA-Z0-9]*)|([-=+_a-zA-Z][-\\/:.=_+a-zA-Z0-9]*\([-+\\/:.=_a-zA-Z0-9]*\)) {
                          yylval.ptr.str=NULL; /*19102004 */
                          my_strdup(300, &yylval.ptr.str, yytext); /* these things are freed after use in expandlabel.y */
                          if(dbg_var>=3) fprintf(errfp, "parselabel(): B_NAME: %s\n", yytext);
                          return B_NAME;
                        }
\[                      {
                          bracket++;
                          if(dbg_var>=3) fprintf(errfp, "parselabel(): open bracket: %s\n", yytext);
                          return yytext[0];
                        }
\]                      {
                          bracket--;
                          if(dbg_var>=3) fprintf(errfp, "parselabel(): close bracket: %s\n", yytext);
                          return yytext[0];
                        }
[^ ]                    {
                          if(dbg_var>=3) fprintf(errfp, "parselabel(): returning character: %s\n", yytext);
                          return yytext[0];
                        }
} /* end <label> */

<rest>{                 /* treat rest of line as a single LINE token */
(.|\n)*                 {
                          yylval.ptr.str=NULL; /*19102004 */
                          my_strdup(301, &yylval.ptr.str, yytext); /* these things are freed after use in expandlabel.y */
                          if(dbg_var>=3) fprintf(errfp, "yylex(): <rest> B_LINE: %s\n",yytext);
                          BEGIN(INITIAL);
                          return B_LINE;
                        }

} /* end rest */
<*><<EOF>>              {
                          BEGIN(INITIAL);
                          return 0;
                        }
%%

